module BRAMInitiator(CLK, RST_N,
            RD_ADDR, RD_RDY,   RD_EN,
            DOUT,    DOUT_RDY, DOUT_EN,
            WR_ADDR, WR_VAL,   WR_EN,
            BRAM_Addr, BRAM_Dout, BRAM_Din,BRAM_Dummy_Enable, 
            BRAM_WEN, BRAM_EN, BRAM_RST, 
            BRAM_CLK); 

   // synopsys template   
   parameter                   addr_width = 1;
   
   input                       CLK;
   input                       RST_N;   

   // Read Port
   // req
   input [addr_width -1 : 0]   RD_ADDR;
   input                       RD_EN;
   output                      RD_RDY;
   // resp
   output [31 : 0]             DOUT;
   output                      DOUT_RDY;
   input                       DOUT_EN;

   // Write Port
   // req
   input [addr_width - 1 : 0]  WR_ADDR;
   input [31 : 0]              WR_VAL;
   input                       WR_EN;

   // BRAM Wires
   output [addr_width - 1 : 0] BRAM_Addr; 
   output [31 : 0]             BRAM_Dout;
   input  [31 : 0]             BRAM_Din; 
   input                       BRAM_Dummy_Enable;
   output [3  : 0]             BRAM_WEN;
   output                      BRAM_EN; 
   output                      BRAM_RST; 
   output                      BRAM_CLK;


   // Assignments
   assign BRAM_CLK = CLK; 
   assign BRAM_EN = RST_N; // disable the BRAM if we are in reset  
   assign BRAM_RST = 1'b0; // Never reset the BRAM.
   assign BRAM_Addr = (WR_EN)?(WR_ADDR):(RD_ADDR);
   assign BRAM_WEN  = {WR_EN,WR_EN,WR_EN,WR_EN};
   assign BRAM_Dout = WR_VAL;
   
   reg                         RD_REQ_MADE;
   
   reg  [1:0] CTR;
   
   /*always@(BRAM_Din)
     $display("BRAMInitiator.v BRAM_Din: %x",BRAM_Din); 
   */

   FIFO2#(.width(32)) q(.RST_N(RST_N),
                        .CLK(CLK),
                        .D_IN(BRAM_Din),
                        .ENQ(RD_REQ_MADE),
                        .DEQ(DOUT_EN),
                        .CLR(1'b0),
                        .D_OUT(DOUT),
                        .FULL_N(),
                        .EMPTY_N(DOUT_RDY));

   assign RD_RDY = (CTR > 0);

   always@(posedge CLK)
     begin       
       if (!RST_N)
         begin  
           CTR <= 2;
         end
       else
         begin
           /*if(RD_EN)
             $display("BRAMInitiator: RD_EN");
           if(WR_EN)
             $display("BRAMInitiator: WR_EN");
           */

           RD_REQ_MADE <= RD_EN;
          
           CTR <= (RD_EN) ?
                    (DOUT_EN) ? CTR : CTR - 1 :
                    (DOUT_EN) ? CTR + 1 : CTR; 
         end
     end // always@ (posedge CLK)

endmodule