// Copyright 2000--2006 Bluespec, Inc.  All rights reserved.

`ifdef BSV_ASSIGNMENT_DELAY
`else
`define BSV_ASSIGNMENT_DELAY
`endif


// A synchronization module for resets.   Output resets are held for
// RSTDELAY cycle, RSTDELAY >= 0.   Both assertion and deassertions is
// synchronized to the clock.
// Note the Bluespec library does not allow RSTDELAY to be 0, since
// value allows OUT_RST_N to be asserted for less than 1 clock cycle.
module SyncReset (
                  IN_RST_N,
                  CLK,
                  OUT_RST_N
                  );
  
   parameter          RSTDELAY = 2  ; // Width of reset shift reg
   
   input              CLK ;
   input              IN_RST_N ;
   output             OUT_RST_N ;

   reg [RSTDELAY:0]   reset_hold ;
   
   assign  OUT_RST_N = reset_hold[RSTDELAY] ;

   always @( posedge CLK )      // reset is read synchronous with clock
     begin
        if ( ! IN_RST_N )
           begin
              reset_hold <= `BSV_ASSIGNMENT_DELAY 'b0 ;
           end
        else
          begin
             reset_hold <= `BSV_ASSIGNMENT_DELAY ( reset_hold << 1 ) | 'b1 ;             
          end
     end // always @ ( posedge CLK )
   
`ifdef BSV_NO_INITIAL_BLOCKS
`else // not BSV_NO_INITIAL_BLOCKS
   // synopsys translate_off
   initial
     begin
        // initialize out of reset forcin the designer to do one.
        reset_hold = {(RSTDELAY + 1) {1'b1 }} ;
     end
   // synopsys translate_on
`endif // BSV_NO_INITIAL_BLOCKS
   
endmodule // SyncReset
