#include <stdio.h>
#include "cLib.h"
#include "timer.h"
// #include "matrixPointers.h"

#define MATRIX_A            0x00000000
#define MATRIX_B            0x00100000
#define MATRIX_C            0x00200000
#define PROGRAM             0x00300000

#define INST_FULL_ADDR      0x00000000
#define INST_FIFO_ADDR      0x00000001
#define DATA_IN_FULL_ADDR   0x00000002
#define DATA_IN_FIFO_ADDR   0x00000003
#define DATA_OUT_EMPTY_ADDR 0x00000004
#define DATA_OUT_FIFO_ADDR  0x00000005
#define STORE_ADDR          0x00000006

#define ZERO                0x00000000
#define LOAD                0x00000001
#define LOAD_MUL            0x00000002
#define MUL                 0x00000003
#define STORE               0x00000004
#define LOADADDR            0x00000005
#define STOREADDR           0x00000006
#define SET_ROW_SIZE        0x00000007

#define BLOCK_SIZE          16
#define BLOCK_SIZE_64		 64


inline void zero()
{
    while(putInstruction(0, ZERO) == -1)
    {
    }
}

inline void load(int i, int j, int size)
{
    while(putInstruction(MATRIX_A + i*BLOCK_SIZE*size + j*BLOCK_SIZE, LOADADDR) == -1)
    {
    }

    while(putInstruction(0, LOAD))
    {
    }
}

inline void loadMul(int i, int j, int size)
{
    while(putInstruction(MATRIX_B + i*BLOCK_SIZE*size + j*BLOCK_SIZE, LOADADDR) == -1)
    {
    }

    while(putInstruction(0, LOAD_MUL))
    {
    }
}

inline void store(int i, int j, int size)
{
   int loopy;

    while(putInstruction(MATRIX_C + i*BLOCK_SIZE*size + j*BLOCK_SIZE, STOREADDR) == -1)
    {
    }

    while(putInstruction(0, STORE))
    {
    }
	 
	 while(getResponse() < 0)
    {
	    loopy++;
       if(loopy == 300000000)
       {		 
	        //xil_printf("Store died: %d %d\r\n", i, j);
	    }
	 }
}

inline void setRowSize(int rowSize)
{
    while(putInstruction(rowSize, SET_ROW_SIZE) == -1)
    {
    }
}

int multiplyFunction(int size, int logSize)
{
    int blockNum = size/BLOCK_SIZE;

    int storeTotal = blockNum*blockNum;

    int i, j, k;

    initialize();

    setRowSize(logSize);

    for(i = 0; i < blockNum; i++)
    {
        for(j = 0; j < blockNum; j++)
        {
            zero();
            for(k = 0; k < blockNum; k++)
            {
                load(k, j, size);
                loadMul(i, k, size);
            }
            store(i, j, size);
        }
    }
}

int multiplyFunction64(volatile int* program, int size, int logSize){
	
#define createSec(m, i, j, x) \
            (m + i*BlockSize*Size*4 + (FU_Number*j+x)*BlockSize*4)

#define createPrim(m, i, j) \
            (m + i*BlockSize*Size*4 + j*BlockSize*4)

    int Size = size;
    int BlockSize = BLOCK_SIZE_64;
    UInt64 a = (UInt64)0x0000000000000000;
    UInt64 b = (UInt64)0x0000000000400000;
    UInt64 c = (UInt64)0x0000000000800000;
	 volatile int* startptr = program;
	 volatile int* endptr = program;

    // we are only using one functional unit
    int FU_Number = (int) 1;

    int BlockNum = Size/BlockSize;
    int LogSize = logSize;

    int BigBlockNum = BlockNum/FU_Number;
    int BigBlockRest = BlockNum%FU_Number;
    //xil_printf("%d %d %d\r\n", BlockNum, BigBlockNum, FU_Number);
    int response;
    int stores_received = 0;
	 int number_stores = (size/64)*(size/64);

    UInt64 inst;    
  
    int i;
    int j;
    int k;
    int f=0;
    int top;
	 int bot;
    
	 int num_insts = 0;
	 int num_sent = 0;
	 
    inst = createSetRowSizeInstruction(LogSize);
    top = (inst >> 32) & 0xffffffff;
    bot = inst & 0xffffffff;
	 num_insts++;
    //xil_printf("BOOGA!\r\n", num_insts);
    *(endptr++) = top;
	 *(endptr++) = bot;
	 
	 for(i = 0; i < BlockNum; i=i+1)
    {
        for(j = 0; j < BlockNum; j=j+1)
        {
	     inst = createArithmeticInstruction(All_FU_Mask, Zero);
		  num_insts++;
        //xil_printf("BOOGA! %d\r\n", num_insts);
        top = (inst >> 32) & 0xffffffff;
		  bot = inst & 0xffffffff;
    
		  *(endptr++) = top;
    	  *(endptr++) = bot;
        for(k = 0; k < BlockNum; k=k+1)
            {
		         inst = createLoadInstruction((((int)1)<<f), B, createSec(b, k, j, f)); 
		         top = (inst >> 32) & 0xffffffff;
					bot = inst & 0xffffffff;
		         num_insts++;
               //xil_printf("BOOGA! %d\r\n", num_insts);
               *(endptr++) = top;
    	         *(endptr++) = bot;
               
		         inst = createLoadInstruction(All_FU_Mask, A, createPrim(a, i, k));
		         top = (inst >> 32) & 0xffffffff;
					bot = inst & 0xffffffff;
		         num_insts++;
               //xil_printf("BOOGA! %d\r\n", num_insts);
               *(endptr++) = top;
    	         *(endptr++) = bot;

               inst = createArithmeticInstruction(All_FU_Mask, MultiplyAddAccumulate); 
		         top = (inst >> 32) & 0xffffffff;
					bot = inst & 0xffffffff;
		         num_insts++;
               //xil_printf("BOOGA! %d\r\n", num_insts);
               *(endptr++) = top;
    	         *(endptr++) = bot;
            }
				
            //for(f = 0; f < FU_Number; f=f+1)
            //{
            //  int i = 0;
              inst = createStoreInstruction(f, C, createSec(c, i, j, f));
  	           top = (inst >> 32) & 0xffffffff;
				  bot = inst & 0xffffffff;
              
		        num_insts++;
              //xil_printf("FINAL BOOGA! %d\r\n", num_insts);
              *(endptr++) = top;
    	        *(endptr++) = bot;
         }
	}	
	top = *startptr;
	bot = *(startptr + 1);				  
   
	while(startptr != endptr){
	    //
		 //if (num_sent < 18) {
         if (putInstruction(top,bot) >= 0){
	      //xil_printf("top bottom %x %x\r\n", top, bot);
		      //xil_printf("Sending Inst %x %x\r\n", startptr, endptr);
			   startptr +=2;
			   top = *startptr;
	         bot = *(startptr + 1);
            //num_sent++;			 
		   }
		 //}
	    //print("try response\r\n");
  	    response = getResponse();
	    if(response >= 0){
	       if(response == 0){
		     print("Beat\r\n");
		    }
		    else{
            stores_received++;
		      //print("got store response\r\n");
				//num_sent = 0;
            //if( stores_received == number_stores) break;
		    }
	    }
    }//end while
		 
		 //print("done sending inst\r\n");
		 while (stores_received != number_stores) {
  	    //xil_printf("stores received: %d\r\n", stores_received);
		 response = getResponse();
	    if(response >= 0){
	       if(response == 0){
		     print("Beat\r\n");
		    }
		    else{
            stores_received++;
		      //print("got store response\r\n");
            if( stores_received == number_stores) break;
		    }
	    }
 }
}