// Global imports
import GetPut::*;
import FIFO::*;
import Vector::*;

// Local imports
import Types::*;
import Parameters::*;

import PLBMasterWires::*;
import BRAMTargetWires::*;
import BRAMInitiatorWires::*;

/* This is the top level interface. We only have wires going to the 
   plb bus and the DSOCM which implements the BRAM interface. 
*/

interface Multiplier;

  interface PLBMasterWires                  plbMasterWires;
  interface BRAMInitiatorWires#(Bit#(14))   bramInitiatorWires;
endinterface

interface MultiplierBackup;

  interface BRAMInitiatorWires#(Bit#(14))   plbBRAMWires;
  interface BRAMInitiatorWires#(Bit#(14))   bramInitiatorWires;
endinterface

interface Controller;

  interface Put#(Instruction)          instructionInput;
  interface Get#(PLBMasterCommand)     plbMasterCommandOutput;
  interface Get#(MemorySwitchCommand)  memorySwitchCommandOutput;
  interface Get#(FUNetworkCommand)     fuNetworkCommandOutput;
  interface Vector#(FunctionalUnitNumber, Get#(FunctionalUnitCommand)) functionalUnitCommandOutputs;

endinterface


interface FunctionalUnit#(numeric type t);
  interface Put#(FunctionalUnitCommand) functionalUnitCommandInput;
  interface FUNetworkLink               link;
  interface Put#(ComplexWord)           switchInput;
  interface Get#(ComplexWord)           switchOutput;
endinterface


interface MemorySwitch;
  interface Put#(ComplexWord)              plbMasterComplexWordInput;
  interface Get#(ComplexWord)              plbMasterComplexWordOutput;
  interface Put#(MemorySwitchCommand)      memorySwitchCommandInput;
  interface Vector#(FunctionalUnitNumber, 
                    Get#(ComplexWord))     functionalUnitComplexWordOutputs;
  interface Vector#(FunctionalUnitNumber, 
                    Put#(ComplexWord))     functionalUnitComplexWordInputs;

  interface Get#(PPCMessage)               getInstructionOutput;
endinterface


interface PLBMaster;
  interface Put#(ComplexWord) wordInput;
  interface Get#(ComplexWord) wordOutput;
  interface Put#(PLBMasterCommand) plbMasterCommandInput;
  interface PLBMasterWires   plbMasterWires;
endinterface

interface PLBMasterBackup;
  interface Put#(ComplexWord) wordInput;
  interface Get#(ComplexWord) wordOutput;
  interface Put#(PLBMasterCommand) plbMasterCommandInput;
  interface BRAMInitiatorWires#(Bit#(14))  plbBRAMWires;
endinterface

  
interface Feeder;

  interface Put#(PPCMessage)  ppcMessageInput;
  interface Get#(Instruction) ppcInstructionOutput;
  interface BRAMInitiatorWires#(Bit#(14))  bramInitiatorWires;

endinterface

interface FUNetwork;

  interface Put#(FUNetworkCommand) fuNetworkCommandInput;

endinterface


interface FUNetworkLink;

  interface Get#(ComplexWord) a_out;
  interface Get#(ComplexWord) b_out;
  interface Get#(ComplexWord) c_out;
  
  interface Put#(ComplexWord) a_in;
  interface Put#(ComplexWord) b_in;
  interface Put#(ComplexWord) c_in;

endinterface

