// Global Imports
import FIFO::*;
import Vector::*;
import GetPut::*;

// Local Imports
`include "Common.bsv"

//Simple controller, FIFOs everywhere.
//Works, but room for improvement.

(* synthesize *)
module [Module] mkSimpleController (Controller);

  Vector#(FunctionalUnitNumber, FIFO#(FunctionalUnitCommand)) fuQs <- replicateM(mkFIFO);
  
  FIFO#(PLBMasterCommand)     plbQ <- mkFIFO();
  FIFO#(MemorySwitchCommand)  switchQ <- mkFIFO();
  FIFO#(FUNetworkCommand)     funetQ <- mkFIFO();
  FIFO#(Instruction)          insQ <- mkFIFO();
  
  
  //Note: The CAN_FIRE of this will be bad with aggressive-conditions
  
  rule decode (True);
    debug(controllerDebug, $display("Controller: processing instruction: %h", insQ.first()));  
    let ins = insQ.first();
    insQ.deq();
    
    // Some debug
    case (insQ.first()) matches
      tagged ArithmeticInstruction .i:   //{.fus, .op}
         debug(controllerDebug, $display("Controller: processing Arithmetic instruction: op: %s fus: %b",
                                showFunctionalUnitOp(i.op),
                                i.fus));
      tagged LoadInstruction .i:         //{.fus, .regName, .addr}
         debug(controllerDebug, $display("Controller: processing Load instruction: fus: %b reg: %s addr: %h",
                                i.fus,
                                showReg(i.regName),
                                i.addr));

      tagged StoreInstruction .i:        //{.fu, .regName, .addr}
         if(valueof(FunctionalUnitNumber) > 1)
           begin
             debug(controllerDebug, $display("Controller: processing Store instruction: fu: %d reg: %s addr: %h",
                                  i.fu,
                                  showReg(i.regName),
                                  i.addr));
           end
         else
           begin
             debug(controllerDebug, $display("Controller: processing Store instruction: fu: %d reg: %s addr: %h",
                                  0,
                                  showReg(i.regName),
                                  i.addr));
           end   

      tagged ForwardInstruction .i:      //{.fuSrc, .regSrc, .fuDests, .regDest}
         if(valueof(FunctionalUnitNumber) > 1)
           begin
         debug(controllerDebug, $display("Controller: processing Forward instruction: fuSrc: %d regSrc: %s fuDests: %b regDest: %s",
                                i.fuSrc,
                                showReg(i.regSrc),
                                i.fuDests,
                                showReg(i.regDest)));
           end
         else
           begin
         debug(controllerDebug, $display("Controller: processing Forward instruction: fuSrc: %d regSrc: %s fuDests: %b regDest: %s",
                                0,
                                showReg(i.regSrc),
                                i.fuDests,
                                showReg(i.regDest)));
           end
      tagged SetRowSizeInstruction .sz:
         debug(controllerDebug, $display("Controller: processing RowSize instruction: log size: %d",
                                         sz));
      default:
         $display("Controller: ERROR, illegal instruction: %h", insQ.first());
    endcase


     //Handle FU commands
    
    let destFUs = case (ins) matches
      tagged ArithmeticInstruction .i:   //{.fus, .op}
        return i.fus;
      tagged LoadInstruction .i:         //{.fus, .regName, .addr}
        return i.fus;
      tagged StoreInstruction .i:        //{.fu, .regName, .addr}
        return oneHot(i.fu);
      tagged ForwardInstruction .i:      //{.fuSrc, .regSrc, .fuDests, .regDest}
        return i.fuDests; //The src is handled separately
      tagged SetRowSizeInstruction .sz:
        return 0;
    endcase;
  
    let fu_cmd =  case (ins) matches
      tagged ArithmeticInstruction .i:   //{.fus, .op}
        return tagged Op i.op;
      tagged LoadInstruction .i:         //{.fus, .regName, .addr}
        return tagged Load i.regName;
      tagged StoreInstruction .i:        //{.fus, .regName, .addr}
        return tagged Store i.regName;
      tagged ForwardInstruction .i:      //{.fuSrc, .regSrc, .fuDests, .regDest}
        return tagged ForwardDest i.regDest;
      tagged SetRowSizeInstruction .sz:
        return ?;
    endcase;
    
    Bool isFwd =  case (ins) matches
      tagged ForwardInstruction .i: 
        return True;
      default:
        return False;
    endcase;
    
    let fwd_src = case (ins) matches
      tagged ForwardInstruction .i: 
        return i.fuSrc;
      default:
        return ?;
    endcase;
    
    let fwd_cmd = case (ins) matches
      tagged ForwardInstruction .i: 
        return ForwardSrc (i.regSrc);
      default:
        return ?;
    endcase;
    
    //Handle Switch Commands
    
    Bool isSwitch = case (ins) matches
      tagged LoadInstruction .i: 
        return True;
      tagged StoreInstruction .i:
        return True;
      default:
        return False;
    endcase;
    
    MemorySwitchCommand sw_cmd = case (ins) matches
      tagged LoadInstruction .i: 
        return tagged LoadToFUs i.fus;
      tagged StoreInstruction .i:
        return tagged StoreFromFU i.fu;
      default:
        return ?;
    endcase;
    
    //Handle PLB Commands
    
    Bool isPLB = case (ins) matches
      tagged SetRowSizeInstruction .sz:
        return True;
      tagged LoadInstruction .i:
        return True;
      tagged StoreInstruction .i:
        return True;
      default:
        return False;
    endcase;
    
    PLBMasterCommand plb_cmd = case (ins) matches
      tagged LoadInstruction .i: 
        return tagged LoadPage truncate(i.addr>>2);
      tagged StoreInstruction .i:
        return tagged StorePage truncate(i.addr>>2);
      tagged SetRowSizeInstruction .sz:
        return tagged RowSize sz;
      default:
        return ?;
    endcase;

    //Handle FUNetwork Commands
    
    Bool isFUNet = case (ins) matches
      tagged ForwardInstruction .i:      //{.fuSrc, .regSrc, .fuDests, .regDest}
        return True;
      default:
        return False;
    endcase;
    
    FUNetworkCommand funet_cmd = case (ins) matches
      tagged ForwardInstruction .i:
        return FUNetworkCommand {fuSrc: i.fuSrc, regSrc: i.regSrc, fuDests: i.fuDests, regDest: i.regDest};
      default:
        return ?;
      endcase;
    

    //Actually do the enqueues
    
    for (Integer x = 0; x < valueOf(FunctionalUnitNumber); x = x + 1)
      if (unpack(destFUs[x]))
      begin
          fuQs[x].enq(fu_cmd);
	  if (x != 0)
	    $display("WARNING: Controller sending commnad to an FU other than 0!");
      end
      else if (isFwd && fwd_src == fromInteger(x))
      begin
	  fuQs[x].enq(fwd_cmd); //Add the forward src command
	  if (x != 0)
	    $display("WARNING: Controller sending fwdSrc to an FU other than 0!");
      end
  
    if (isSwitch)
      switchQ.enq(sw_cmd);
      
    if (isPLB) 
      begin
        debug(controllerDebug, $display("Controller: enqueuing PLB command"));
        plbQ.enq(plb_cmd);
      end

    if (isFUNet)
      funetQ.enq(funet_cmd);
  
  
  endrule
  
  interface instructionInput = fifoToPut(insQ);
  interface plbMasterCommandOutput = fifoToGet(plbQ);
  interface memorySwitchCommandOutput = fifoToGet(switchQ);
  interface fuNetworkCommandOutput = fifoToGet(funetQ);
  interface functionalUnitCommandOutputs = map(fifoToGet, fuQs);


endmodule
